import dapgen
import pandas as pd
import admix
import numpy as np

import dapgen
import pandas as pd
import admix
import numpy as np
from typing import List
import os
import glob
import subprocess
from ._utils import get_dependency


def run(cmd: str, **kwargs):
    """Shortcut for running plink commands

    Parameters
    ----------
    cmd : str
        plink command
    """
    bin_path = get_dependency("plink2")
    add_cmds = [f" --{k.replace('_', '-')} {kwargs[k]}" for k in kwargs]
    cmd += " ".join(add_cmds)
    subprocess.check_call(f"{bin_path} {cmd}", shell=True)


def gwas(
    df_sample_info: pd.DataFrame,
    pheno_col: str,
    out_prefix: str,
    pfile: str = None,
    bfile: str = None,
    family: str = "linear",
    covar_cols: List[str] = None,
    cat_cols: List[str] = None,
    pheno_quantile_normalize=False,
    covar_quantile_normalize=False,
    clean_tmp_file=False,
    **kwargs,
):
    assert family in ["linear", "logistic"], "family must be linear or logistic"
    # only one of pfile or bfile must be provided
    assert (pfile is None) != (bfile is None), "pfile or bfile must be provided"

    # clean up phenotype
    non_nan_index = ~np.isnan(df_sample_info[pheno_col])
    df_pheno = df_sample_info.loc[non_nan_index, :].rename(columns={pheno_col: "trait"})
    pheno_vals = df_pheno.trait.values
    if family == "logistic":
        assert (
            pheno_quantile_normalize == False
        ), "quantile normalization can not be used for logistic regression"
        assert np.all(
            (pheno_vals == 0) | (pheno_vals == 1)
        ), "phenotype values must be 0 or 1 for logistic regression"
        df_pheno["trait"] = df_pheno.trait.astype(int) + 1
    elif family == "linear":
        df_pheno["trait"] = (pheno_vals - pheno_vals.mean()) / pheno_vals.std()
    else:
        raise ValueError("family must be linear or logistic")
    pheno_path = out_prefix + f".plink2_tmp_pheno"

    if bfile is not None:
        # PLINK1 format
        # FID, IID must be in the column
        assert "FID" in df_sample_info.columns
        assert "IID" in df_sample_info.columns
        df_pheno[["FID", "IID", "trait"]].to_csv(
            pheno_path, sep="\t", index=False, na_rep="NA"
        )
        cmds = [f"--bfile {bfile}"]
    elif pfile is not None:
        # PLINK2 format
        df_pheno.index.name = "#IID"
        df_pheno[["trait"]].to_csv(pheno_path, sep="\t", na_rep="NA")
        cmds = [f"--pfile {pfile}"]

    cmds.extend(
        [
            f"--pheno {pheno_path} --no-psam-pheno",
            f"--out {out_prefix}",
        ]
    )
    if family == "linear":
        cmds.append("--linear hide-covar omit-ref")
    elif family == "logistic":
        cmds.append("--logistic hide-covar omit-ref no-firth")
    else:
        raise ValueError("family must be linear or logistic")

    if covar_cols is not None:
        covar_path = out_prefix + f".plink2_tmp_covar"
        cmds.append(f"--covar {covar_path}")
        if bfile is not None:
            # PLINK1 format
            # FID, IID must be in the column
            assert "FID" in df_sample_info.columns
            assert "IID" in df_sample_info.columns
            df_covar = df_sample_info.loc[
                non_nan_index, ["FID", "IID"] + covar_cols
            ].copy()
            df_covar.to_csv(covar_path, sep="\t", index=False, na_rep="NA")
        elif pfile is not None:
            # PLINK2 format
            df_covar = df_sample_info.loc[non_nan_index, covar_cols].copy()
            df_covar.index.name = "#IID"
            df_covar.to_csv(covar_path, sep="\t", na_rep="NA")

    else:
        cmds[-1] += " allow-no-covars"

    if pheno_quantile_normalize:
        cmds.append("--pheno-quantile-normalize")
    if covar_quantile_normalize:
        cmds.append("--covar-quantile-normalize")
    if cat_cols:
        cmds.append(f"--split-cat-pheno omit-most {' '.join(cat_cols)}")

    print("\n".join(cmds))

    run(" ".join(cmds), **kwargs)

    if family == "linear":
        os.rename(out_prefix + ".trait.glm.linear", out_prefix + ".assoc")
    else:
        os.rename(out_prefix + ".trait.glm.logistic", out_prefix + ".assoc")

    if clean_tmp_file:
        for f in glob.glob(out_prefix + ".plink2_tmp_*"):
            os.remove(f)


def clump(
    pfile,
    assoc_path: str,
    out_prefix: str,
    p1: float = 5e-8,
    p2: float = 1e-4,
    r2: float = 0.1,
    kb=3000,
    **kwargs,
):
    """
    Wrapper for plink2 clump
    For now, first need to export to .bed format then perform the clump
    """
    tmp_prefix = out_prefix + ".admix_plink2_clump_tmp"
    # convert to bed
    admix.tools.plink2.run(f"--pfile {pfile} --make-bed --out {tmp_prefix}")

    # perform clump
    admix.tools.plink.clump(
        bfile=tmp_prefix,
        assoc_path=assoc_path,
        out_prefix=out_prefix,
        p1=p1,
        p2=p2,
        r2=r2,
        kb=kb,
        **kwargs,
    )
    # # convert plink2 association to plink1 format ID -> SNP
    # import shutil

    # from_file = open(assoc_path)
    # to_file = open(tmp_prefix + ".assoc", "w")
    # to_file.writelines(from_file.readline().replace("ID", "SNP"))
    # shutil.copyfileobj(from_file, to_file)
    # from_file.close()
    # to_file.close()
    # cmds = [
    #     f"--bfile {tmp_prefix} --clump {tmp_prefix + '.assoc'}",
    #     f"--clump-p1 {p1} --clump-p2 {p2} --clump-r2 {r2} --clump-kb {kb}",
    #     f"--out {tmp_prefix}",
    # ]

    # admix.tools.plink.run(" ".join(cmds))
    # if os.path.exists(tmp_prefix + ".clumped"):
    #     os.rename(tmp_prefix + ".clumped", out_prefix + ".clumped")
    # else:
    #     # no clumped region
    #     # write a comment to the output file
    #     with open(out_prefix + ".clumped", "w") as file:
    #         file.write("# No clumped region")

    for f in glob.glob(tmp_prefix + "*"):
        os.remove(f)


def lift_over(pfile: str, out_prefix: str, chain="hg19->hg38"):
    """Lift over a plink file to another genome. First, use liftover to convert to
    another genome, then use plink2 --sort-vars to sort the variants.

    Parameters
    ----------
    pfile : str
        Path to plink file
    out_prefix : str
        output prefix, <out_prefix>.pgen, <out_prefix>.psam, <out_prefix>.pvar will
        be generated
    chain : str, optional
        lift over direction, by default "hg19->hg38"
    """
    assert chain in ["hg19->hg38", "hg38->hg19"]

    # read the input pgen
    pgen, pvar, psam = dapgen.read_pfile(pfile)
    df_snp = pd.DataFrame(
        {"CHROM": pvar.CHROM.values, "POS": pvar.POS.values}, index=pvar.index.values
    )
    # perform the liftover
    df_lifted = admix.tools.liftover.run(df_snp, chain=chain)
    n_snp1 = len(df_lifted)
    df_lifted = df_lifted[df_lifted.POS != -1]
    n_snp2 = len(df_lifted)

    print(
        f"remove {n_snp1 - n_snp2} / {n_snp1} SNPs"
        f" ({(n_snp1 - n_snp2) / n_snp1 * 100:.2g}%)"
        " for unmapped or ambiguous SNPs"
    )

    # extract the lifted SNPs
    np.savetxt(f"{out_prefix}-tmp.snp", df_lifted.index.values, fmt="%s")
    admix.tools.plink2.run(
        f"--pfile {pfile} --extract {out_prefix}-tmp.snp --sort-vars --make-pgen --out {out_prefix}-tmp"
    )

    # substitute the coordinates obtained from liftover
    pgen, pvar, psam = dapgen.read_pfile(out_prefix + "-tmp")
    assert np.all(pvar.index == df_lifted.index)

    pvar["POS"] = df_lifted["POS"]
    pvar.insert(2, "ID", pvar.index)
    pvar = pvar.reset_index(drop=True).rename(columns={"CHROM": "#CHROM"})
    pvar.to_csv(f"{out_prefix}-tmp.pvar", sep="\t", index=False)

    admix.tools.plink2.run(
        f"--pfile {out_prefix}-tmp --sort-vars --make-pgen --out {out_prefix}"
    )
    # remove the tmp files
    for f in glob.glob(out_prefix + "-tmp*"):
        os.remove(f)


def align_snp(pfile1: str, pfile2: str, out_prefix: str):
    """Given two plink files from different individuals, take the common set of SNPs.
    Matching by (1) chromosome (2) position (3) ref allele (4) alt allele

    This can be useful for software that requires an aligned set of SNPs

    Parameters
    ----------
    pfile1 : str
        plink file 1
    pfile2 : str
        plink file 2
    out_prefix: str
        prefix to the output files
    """
    # Step 1: read pvar and find the common SNPs
    df_snp1 = dapgen.read_pvar(pfile1 + ".pvar").drop_duplicates(
        subset=["CHROM", "POS"]
    )
    df_snp2 = dapgen.read_pvar(pfile2 + ".pvar").drop_duplicates(
        subset=["CHROM", "POS"]
    )

    idx1, idx2, flip = dapgen.align_snp(df_snp1, df_snp2)

    admix.logger.info(
        f"{len(idx1)} / {len(df_snp1)} ({len(idx1) / len(df_snp1) * 100:.3g}%) SNPs extracted in pfile1: {pfile1}"
    )
    admix.logger.info(
        f"{len(idx2)} / {len(df_snp2)} ({len(idx2) / len(df_snp2) * 100:.3g}%) SNPs extracted in pfile2: {pfile2}"
    )

    # Step 2: extract the common SNPs from the two files and fix the allele order
    allele1 = df_snp1.loc[idx1]["REF"]
    allele2 = df_snp2.loc[idx2]["REF"]

    allele1.to_csv(out_prefix + "-tmp.1.refallele", sep="\t", header=False)
    np.savetxt(out_prefix + "-tmp.1.snplist", allele1.index.values, fmt="%s")

    # use sample ref-allele values, but use ref-allele index
    pd.Series(allele1.values, index=allele2.index).to_csv(
        out_prefix + "-tmp.2.refallele", sep="\t", header=False
    )
    np.savetxt(out_prefix + "-tmp.2.snplist", allele2.index.values, fmt="%s")

    # extract common SNPs and align REF ALT
    admix.tools.plink2.run(
        f"--pfile {pfile1} --extract {out_prefix + '-tmp.1.snplist'}"
        f" --ref-allele {out_prefix + '-tmp.1.refallele'} 2 1 --make-pgen --out {out_prefix}.1"
    )
    admix.tools.plink2.run(
        f"--pfile {pfile2} --extract {out_prefix + '-tmp.2.snplist'} "
        f" --ref-allele force {out_prefix + '-tmp.2.refallele'} 2 1 --make-pgen --out {out_prefix}.2"
    )

    # remove the tmp files
    for f in glob.glob(out_prefix + "-tmp*"):
        os.remove(f)


def merge_indiv(pfile1: str, pfile2: str, out_prefix: str):
    """Given two plink files from different individuals, take the common set of SNPs
    and merge them into one data set.

    This is a wrapper for various PLINK functions. Since PLINK2's pmerge function is
    not complete now. We use PLINK1 merge functionality instead. A PLINK1 bfile is
    produced (rather than PLINK2 pfile).

    This can be useful for example when we want to perform a joint PCA for the two
    data sets.

    Parameters
    ----------
    pfile1 : str
        plink file 1
    pfile2 : str
        plink file 2
    out_prefix: str
        prefix to the output files
    """

    admix.tools.plink2.align_snp(
        pfile1=pfile1, pfile2=pfile2, out_prefix=out_prefix + "-tmp"
    )

    # Step 3: unify SNP ID
    admix.tools.plink2.run(
        f"--pfile {out_prefix}-tmp.1 --set-all-var-ids @:#:\$r:\$a --make-bed --out {out_prefix}-tmp.1.fixid"
    )
    admix.tools.plink2.run(
        f"--pfile {out_prefix}-tmp.2 --set-all-var-ids @:#:\$r:\$a --make-bed --out {out_prefix}-tmp.2.fixid"
    )

    # Step 4: make final merged bfile
    admix.tools.plink.run(
        f"--bfile {out_prefix}-tmp.1.fixid --bmerge {out_prefix}-tmp.2.fixid --make-bed --out {out_prefix}"
    )

    # remove the tmp files
    for f in glob.glob(out_prefix + "-tmp*"):
        os.remove(f)


def subset(
    pfile: str,
    out_prefix: str,
    snp_list: List = None,
    indiv_list: List = None,
):
    cmds = [f"--pfile {pfile}", f"--make-pgen --out {out_prefix}", "--memory 16000"]

    if snp_list is not None:
        snplist_path = out_prefix + ".admix-plink2-subset.snplist"
        np.savetxt(snplist_path, snp_list, fmt="%s")

        cmds.append(f"--extract {snplist_path}")

    if indiv_list is not None:
        indivlist_path = out_prefix + ".admix-plink2-subset.indivlist"
        np.savetxt(indivlist_path, indiv_list, fmt="%s")
        cmds.append(f"--keep {indivlist_path}")

    run(" ".join(cmds))

    # clean up
    for f in glob.glob(out_prefix + ".admix-plink2-subset.*"):
        os.remove(f)


# TODO: check difference with plink2.gwas, consider merging these two functions
def assoc(
    bfile: str,
    pheno: pd.DataFrame,
    out_prefix: str,
    covar: pd.DataFrame = None,
    indiv: pd.DataFrame = None,
    snp: pd.DataFrame = None,
):
    """Run plink association test

    Parameters
    ----------
    bfile : str
        plink binary file
    pheno : pd.DataFrame
        phenotype data
    out_prefix : str
        prefix for output files
    covar : pd.DataFrame, optional
        covariate data
    indiv : pd.DataFrame, optional
        individual to subset, should contain columns FID and IID
    snp : pd.DataFrame, optional
        snp to subset, should contain column SNP
    """

    assert indiv is None and snp is None, "indiv and snp are not supported"
    pheno_path = out_prefix + ".tmp_pheno"
    covar_path = out_prefix + ".tmp_covar"
    pheno.to_csv(pheno_path, sep="\t", index=False)
    if covar is not None:
        covar.to_csv(covar_path, sep="\t", index=False)

    cmd = [
        f"--bfile {bfile}",
        f"--out {out_prefix}",
        f"--pheno {pheno_path}",
    ]
    if covar is not None:
        cmd.append("--ci 0.95 --glm omit-ref hide-covar")
        cmd.append(f"--covar {covar_path}")
    else:
        cmd.append("--ci 0.95 --glm omit-ref hide-covar allow-no-covars")

    run(" ".join(cmd))


def prune(
    pfile: str,
    out_prefix: str,
    indep_params: List = None,
    indep_pairwise_params: List = None,
):
    """Run plink2 prune

    For example, indep_params = [200, 5, 1.15], indep-pairwise = [100 5 0.1], then two-step
    pruning will be performed.

    Parameters
    ----------
    pfile : str
        plink binary file
    indep_params : List
        list of parameters for indep
    indep_pairwise_params : List
        list of parameters for indep-pairwise
    out_prefix : str
        prefix for output files
    """

    assert indep_params is None, "indep_params is not supported yet"

    tmp_prefix = out_prefix + ".admix_plink2_prune_tmp"

    assert (indep_params is None) != (
        indep_pairwise_params is None
    ), "only one of indep_params and indep_pairwise_params can be specified"

    # step 1: indep pruning
    cmd = [
        f"--pfile {pfile}",
        f"--out {tmp_prefix}",
    ]

    if indep_params is not None:
        cmd.append(f"--indep {' '.join([str(p) for p in indep_params])}")
    if indep_pairwise_params is not None:
        cmd.append(
            f"--indep-pairwise {' '.join([str(p) for p in indep_pairwise_params])}"
        )

    run(" ".join(cmd))

    cmd = [
        f"--pfile {pfile}",
        f"--extract {tmp_prefix}.prune.in",
        f"--make-pgen --out {out_prefix}",
    ]
    run(" ".join(cmd))

    # clean up
    for f in glob.glob(tmp_prefix + ".*"):
        os.remove(f)


def pca(pfile: str, out_prefix: str, approx: bool = False, **kwargs):
    """Run plink2 pca

    TODO: include more options

    Parameters
    ----------
    pfile : str
        plink binary file
    out_prefix : str
        prefix for output files
    approx: bool
    """

    cmd = [
        f"--pfile {pfile}",
        f"--out {out_prefix}",
    ]
    if approx:
        cmd.append("--pca approx")
    else:
        cmd.append("--pca")
    run(" ".join(cmd), **kwargs)


def subset_hapmap3(pfile: str, build: str, out_prefix: str = None, chrom: int = None):
    """Overlap a pfile with HapMap3 SNPs

    Parameters
    ----------
    pfile : str
        input pfile
    build: str
        hg19 or hg38
    out_prefix: str
        output pfile prefix

    Returns
    -------
    <out_prefix>.{pgen|pvar|psam} will be written to disk if out_prefix is specified
    otherwise, snp_list will be returned

    """
    assert build in ["hg19", "hg38"], "build must be hg19 or hg38"
    try:
        import pyreadr
    except ImportError:
        raise ImportError("pyreadr is not installed. Use pip install pyreadr")
    hm3_snps = pyreadr.read_r(admix.tools.get_cache_data("hapmap3_snps"))[None]
    if build == "hg19":
        hm3_snps = hm3_snps[["chr", "pos"]].rename(
            columns={"chr": "CHROM", "pos": "POS"}
        )
    elif build == "hg38":
        hm3_snps = hm3_snps[["chr", "pos_hg38"]].rename(
            columns={"chr": "CHROM", "pos_hg38": "POS"}
        )
    else:
        raise NotImplementedError

    TOTAL_HM3_SNPS = hm3_snps.shape[0]
    if chrom is not None:
        hm3_snps = hm3_snps[hm3_snps.CHROM == chrom]

    df_snp = dapgen.read_pvar(pfile + ".pvar")
    snp_list = df_snp.reset_index().merge(hm3_snps, on=["CHROM", "POS"])["snp"].values

    admix.logger.info(f"{len(snp_list)}/{TOTAL_HM3_SNPS} SNPs are retained")
    if out_prefix is None:
        return snp_list
    else:
        subset(pfile, out_prefix, snp_list=snp_list)



from numpy import (
    asarray,
    float32,
    float64,
    fromfile,
    int64,
    tril,
    tril_indices_from,
    zeros,
)
from pandas import read_csv
import numpy as np
import re
import dask.array as da
from typing import List, Optional
import xarray as xr
import admix
import dapgen
import os
import pandas as pd


def read_lanc(path: str) -> admix.data.Lanc:
    """Read local ancestry with .lanc format

    Parameters
    ----------
    """
    lanc = admix.data.Lanc(path)
    return lanc


def read_dataset(
    pfile: str,
    lanc_file: str = None,
    snp_info_file: str = None,
    indiv_info_file: str = None,
    n_anc: int = None,
    snp_chunk: int = 1024,
) -> admix.Dataset:
    """
    TODO: support multiple pfile, such as data/chr*
    TODO: support only reading a subset of the individuals

    Read a dataset from a directory.

    pfile.snp_info will also be read and combined with .pvar

    Parameters
    ----------
    pfile: str
        PLINK2 file prefix
    lanc_file: str
        local ancestry file, if not provided, `read_dataset` will attempt to find it
        with <pfile>.lanc
    snp_info_file: str
        SNP info file, if not provided, `read_dataset` will attempt to find it
        with <pfile>.snp_info
    indiv_info_file: str
        individual info file, if not provided, `read_dataset` will attempt to find it
        with <pfile>.indiv_info
    n_anc: int
        number of ancestries, if not provided, `read_dataset` will attempt to infer from
        the local ancestry file
    snp_chunk: int
        chunk size for reading the SNP info file (default: 1024)

    Returns
    -------
    Dataset
    """

    # infer local ancestry file
    if lanc_file is None:
        if os.path.exists(pfile + ".lanc"):
            lanc_file = pfile + ".lanc"
    if lanc_file is not None:
        lanc = admix.io.read_lanc(lanc_file).dask(snp_chunk=snp_chunk)
    else:
        lanc = None

    # infer SNP info file
    if snp_info_file is None:
        if os.path.exists(pfile + ".snp_info"):
            snp_info_file = pfile + ".snp_info"

    if indiv_info_file is None:
        if os.path.exists(pfile + ".indiv_info"):
            indiv_info_file = pfile + ".indiv_info"

    geno, pvar, psam = dapgen.read_pfile(pfile, phase=True, snp_chunk=snp_chunk)

    dset = admix.Dataset(geno=geno, lanc=lanc, snp=pvar, indiv=psam, n_anc=n_anc)

    if snp_info_file is not None:
        df_snp_info = pd.read_csv(snp_info_file, index_col=0, sep="\t")
        assert (
            len(set(dset.snp.columns) & set(df_snp_info.columns)) == 0
        ), "SNP info file columns must not overlap with dset columns"
        dset._snp = pd.merge(
            dset.snp,
            df_snp_info.reindex(dset.snp.index),
            left_index=True,
            right_index=True,
        )

    if indiv_info_file is not None:
        df_indiv_info = pd.read_csv(
            indiv_info_file,
            index_col=0,
            sep="\t",
            low_memory=False,
        )
        assert (
            len(set(dset.indiv.columns) & set(df_indiv_info.columns)) == 0
        ), "there should be no intersection between dest.indiv.columns and indiv_info.columns"
        dset._indiv = pd.merge(
            dset.indiv,
            df_indiv_info.reindex(dset.indiv.index),
            left_index=True,
            right_index=True,
        )
    return dset


def read_vcf(path: str, region: str = None, samples: List[str] = None) -> Optional[xr.Dataset]:
    """read vcf file and form xarray.Dataset

    Parameters
    ----------
    path : str
        path to vcf file
    region : str, optional
        region to read, passed to scikit-allel, by default None

    Returns
    -------
    xarray.Dataset
        xarray.Dataset, if no snps in region, return None
    """
    import allel
    import xarray as xr

    vcf = allel.read_vcf(
        path,
        region=region,
        samples=samples,
        fields=["samples", "calldata/GT", "variants/*"],
    )
    if vcf is None:
        return None

    gt = vcf["calldata/GT"]
    assert (gt == -1).sum() == 0

    # used to convert chromosome to int
    chrom_format_func = np.vectorize(lambda x: int(x.replace("chr", "")))
    dset = xr.Dataset(
        data_vars={
            "geno": (("indiv", "snp", "ploidy"), da.from_array(np.swapaxes(gt, 0, 1))),
        },
        coords={
            "snp": vcf["variants/ID"].astype(str),
            "indiv": vcf["samples"].astype(str),
            "CHROM": (
                "snp",
                chrom_format_func(vcf["variants/CHROM"]),
            ),
            "POS": ("snp", vcf["variants/POS"].astype(int)),
            "REF": ("snp", vcf["variants/REF"].astype(str)),
            "ALT": ("snp", vcf["variants/ALT"][:, 0].astype(str)),
            "R2": ("snp", vcf["variants/R2"].astype(float)),
            "MAF": ("snp", vcf["variants/MAF"].astype(float)),
        },
    )
    return dset


def read_digit_mat(path: str, filter_non_numeric: bool = False, nrows: int = None):
    """
    Read a matrix of integer with [0-9], and with no delimiter.

    Parameters
    ----------
    path : str
        path to the matrix file
    filter_non_numeric : bool, optional
        whether to filter out non-numeric characters, by default False
    nrows : int, optional
        number of rows to read, by default None

    Returns
    -------
    np.ndarray
        matrix of integer
    """
    if nrows is None:
        if filter_non_numeric:
            with open(path) as f:
                mat = np.array(
                    [
                        np.array([int(c) for c in re.sub("[^0-9]", "", line.strip())])
                        for line in f.readlines()
                    ],
                    dtype=np.int8,
                )
        else:
            with open(path) as f:
                mat = np.array(
                    [np.array([int(c) for c in line.strip()]) for line in f.readlines()],
                    dtype=np.int8,
                )
    else:
        assert filter_non_numeric is False
        mat = np.array(
            [
                np.array([int(c) for c in line.item()])
                for line in pd.read_csv(path, nrows=nrows, header=None).values
            ],
            dtype=np.int8,
        )
    return mat


def read_gcta_grm(file_prefix) -> dict:
    """
    Reads the GRM from a GCTA formated file.

    Parameters
    ----------
    file_prefix : str
        The prefix of the GRM to be read.

    Returns
    -------
    dict
        A dictionary with the GRM values.
        - grm: GRM matrix
        - df_id: ids of the individuals
        - n_snps: number of SNP

    """

    bin_file = file_prefix + ".grm.bin"
    N_file = file_prefix + ".grm.N.bin"
    id_file = file_prefix + ".grm.id"

    df_id = read_csv(id_file, sep="\t", header=None, names=["sample_0", "sample_1"])
    n = df_id.shape[0]
    k = asarray(fromfile(bin_file, dtype=float32), float64)
    n_snps = asarray(fromfile(N_file, dtype=float32), int64)

    K = zeros((n, n))
    K[tril_indices_from(K)] = k
    K = K + tril(K, -1).T
    return {
        "grm": K,
        "df_id": df_id,
        "n_snps": n_snps,
    }


def read_rfmix(
    path: str,
    df_snp: pd.DataFrame,
    df_indiv: pd.DataFrame,
):
    """
    Assign local ancestry to a dataset. 2-way, 3-way admixture are all supported.

    Parameters
    ----------
    lanc_file: str
        Path to local ancestry data.
    geno: xr.DataArray
        genotype matrix
    df_snp: pd.DataFrame
        SNP data frames

    Returns
    -------
    lanc: da.Array
        Local ancestry array
    """

    # assign local ancestry
    df_rfmix = pd.read_csv(path, sep="\t", skiprows=1)
    assert np.unique(df_rfmix["#chm"]).shape[0] == 1, "rfmix can only contain one chromosome"
    assert np.unique(df_snp["CHROM"]).shape[0] == 1, "df_snp can only contain one chromosome"
    assert set(df_rfmix["#chm"].values) == set(
        df_snp["CHROM"].values
    ), "rfmix and df_snp must contain the same chromosome"
    # read the 2 haplotype, the value of those haplotype corresponds to the local ancestry
    lanc0 = df_rfmix.loc[:, df_rfmix.columns.str.endswith(".0")].rename(columns=lambda x: x[:-2])
    lanc1 = df_rfmix.loc[:, df_rfmix.columns.str.endswith(".1")].rename(columns=lambda x: x[:-2])
    assert (
        np.any([col.endswith(".2") for col in df_rfmix.columns]) == False
    ), "There are columns with .2 in the rfmix file (not expected), please raise an issue"

    lanc = lanc0.astype(str) + lanc1.astype(str)

    df_rfmix_info = df_rfmix.iloc[:, 0:3].copy()
    # extend local ancestry to two ends of chromosomes if necessary
    df_rfmix_info.loc[0, "spos"] = min(df_snp["POS"][0] - 1, df_rfmix_info.loc[0, "spos"])
    df_rfmix_info.loc[len(df_rfmix_info) - 1, "epos"] = max(
        df_snp["POS"][-1] + 1, df_rfmix_info.loc[len(df_rfmix_info) - 1, "epos"]
    )

    assert np.all(df_indiv.index == lanc.columns)

    n_indiv = len(df_indiv)
    n_snp = len(df_snp)

    rfmix_break_list = np.zeros(df_rfmix_info.shape[0], dtype=int)
    # [start, stop) of SNPs for each rfmix break points
    # find the RFmix break points in coordinates of SNP location
    chunk_stop = 0
    for chunk_i, chunk in df_rfmix_info.iterrows():
        chunk_mask = np.where((chunk.spos <= df_snp["POS"]) & (df_snp["POS"] < chunk.epos))[0]
        if len(chunk_mask) > 0:
            chunk_stop = chunk_mask[-1]
        else:
            # use the previous chunk stop if no SNP in this chunk
            admix.logger.warning(f"No SNP found in {chunk_i}-th region of RFmix")
            chunk_stop = chunk_stop
        rfmix_break_list[chunk_i] = chunk_stop

    # find break points in the data
    chunk_pos, indiv_pos = np.where(lanc.iloc[1:, :].values != lanc.iloc[:-1, :].values)
    # convert to SNP positions
    snp_pos = rfmix_break_list[chunk_pos]
    values = lanc.values[chunk_pos, indiv_pos]

    # append values at the end of the chromosomes
    snp_pos = np.concatenate([snp_pos, [n_snp - 1] * n_indiv])
    indiv_pos = np.concatenate([indiv_pos, np.arange(n_indiv)])
    values = np.concatenate([values, lanc.iloc[-1].values])

    # snp_pos, indiv_pos, values are now triples of break points

    break_list = []
    value_list = []
    # convert to .lanc format
    for indiv_i in range(n_indiv):
        indiv_mask = indiv_pos == indiv_i
        # +1 because .lanc denote the [start, stop) of the break points
        indiv_snp_pos, unique_mask = np.unique(snp_pos[indiv_mask] + 1, return_index=True)
        indiv_values = values[indiv_mask][unique_mask]
        break_list.append(indiv_snp_pos.tolist())
        value_list.append(indiv_values.tolist())

    return admix.data.Lanc(breaks=break_list, values=value_list)


def read_joint_pca(pca_prefix: str, ref_pfile: str):
    """
    Read joint PCA results.

    Parameters
    ----------
    ref_pfile : str
        reference panel pfile prefix
    pca_prefix : str
        joint pca results prefix. {pca_prefix}.eigenvec, {pca_prefix}.eigenval
        will be read

    Returns
    -------
    df_pca : pd.DataFrame
        PCA results
    eigenval : np.ndarray
        eigenvalues
    """
    df_pca = (
        pd.read_csv(f"{pca_prefix}.eigenvec", delim_whitespace=True)
        .set_index("IID")
        .drop(columns=["#FID"])
    )

    with open(f"{pca_prefix}.eigenval") as f:
        eigenval = np.array([float(l.strip()) for l in f.readlines()])

    df_pop = pd.read_csv(ref_pfile + ".psam", delim_whitespace=True, index_col=0)
    assert (
        "SuperPop" in df_pop.columns and "Population" in df_pop.columns
    ), f"SuperPop and Population columns are required in the reference panel {ref_pfile}.psam file"
    
    df_pca["SUPERPOP"] = df_pop["SuperPop"].reindex(df_pca.index)
    df_pca["POP"] = df_pop["Population"].reindex(df_pca.index)
    # make sure SAMPLE is not in SUPERPOP or POP
    assert "SAMPLE" not in df_pca["SUPERPOP"].unique()
    assert "SAMPLE" not in df_pca["POP"].unique()
    df_pca["SUPERPOP"] = df_pca["SUPERPOP"].fillna("SAMPLE")
    df_pca["POP"] = df_pca["POP"].fillna("SAMPLE")

    # reorder SAMPLE to the bottom of the data frame
    df_pca = pd.concat([df_pca[df_pca.POP != "SAMPLE"], df_pca[df_pca.POP == "SAMPLE"]])
    return df_pca, eigenval


import pandas as pd
from typing import List
import admix
import numpy as np


def convert_dummy(df: pd.DataFrame, cols: List[str] = None) -> pd.DataFrame:
    """
    Convert dummy variables to categorical variables for each column in df.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to convert
    cols : List[str], optional
        Columns to convert, by default None (columns are selected automatically)

    Returns
    -------
    pd.DataFrame
        Converted dataframe
    """
    if cols is None:
        cols = list(set(df.columns) - set(df._get_numeric_data().columns))

    added_cols = []
    df = df.copy()
    for col in cols:
        # create study dummies variables
        dummies = pd.get_dummies(df[col], drop_first=True)
        dummies.columns = [f"{col}_{s}" for s in dummies.columns]
        dummies.loc[df[col].isnull(), dummies.columns] = np.nan
        added_cols.extend(dummies.columns)
        df = pd.concat([df, dummies], axis=1)
        df = df.drop(columns=[col])
    if len(added_cols) > 0:
        admix.logger.info(
            f"Detected categorical columns: {','.join(cols)}, "
            f"and dded dummy variables: {','.join(added_cols)}"
        )
    return df


import admix
import pandas as pd
import numpy as np
from typing import List
from ._utils import log_params


# def plot_pca(
#     pfile: str,
#     pca: str,
#     out: str,
#     label_col: str = None,
#     x: str = "PC1",
#     y: str = "PC2",
# ):
#     """Plot PCA results to a file

#     Parameters
#     ----------
#     pfile : str
#         pfile
#     label_col : str
#         column in .psam file
#     pca : str
#         path to the pca file
#     out : str
#         path to the output file
#     x : str
#         x-axis (default PC1)
#     y : str
#         y-axis (default PC2)
#     """
#     log_params("plot-pca", locals())

#     import matplotlib.pyplot as plt
#     import matplotlib as mpl

#     mpl.style.use("classic")

#     df_psam = dapgen.read_psam(pfile + ".psam")
#     df_pca = pd.read_csv(pca, delim_whitespace=True, index_col=0)
#     assert np.all(df_psam.index == df_pca.index)

#     df_plot = pd.merge(df_psam, df_pca, left_index=True, right_index=True)

#     fig, ax = plt.subplots(figsize=(2.5, 2.5))
#     admix.plot.pca(df_plot, x=x, y=y, label_col=label_col)

#     # make xticklabels and yticklabels smaller
#     ax.tick_params(axis="x", labelsize=6)
#     ax.tick_params(axis="y", labelsize=6)
#     # make legend font smaller
#     ax.legend(fontsize=8)
#     plt.savefig(out, bbox_inches="tight", dpi=300)


def plot_joint_pca(
    ref_pfile: str,
    pca_prefix: str,
    out: str,
    sample_alpha: float = 0.1,
    figsize=(8.5, 4),
    x="PC1",
    y="PC2",
):
    """Plot individuals on a joint PC plot assuming a joint pca of
    (1) reference dataset (2) sample dataset
    has been performed.

    Parameters
    ----------
    ref_pfile : str
        reference panel pfile prefix
    pca_prefix : str
        joint pca results prefix. {pca_prefix}.eigenvec, {pca_prefix}.eigenval
        will be read
    out : str
        output prefix. {out}.XX, {out}.YY will be produced
    sample_alpha : float
        transparency for sample dots
    figsize : tuple
        figure size (width, height), default (8, 4)
    x : str
        x-axis (default PC1)
    y : str
        y-axis (default PC2)
    """
    log_params("plot-joint-pca", locals())
    import matplotlib.pyplot as plt

    df_plot, eigenval = admix.io.read_joint_pca(
        ref_pfile=ref_pfile, pca_prefix=pca_prefix
    )
    assert set([x, y]).issubset(
        df_plot.columns
    ), f"{x} and {y} must be in the columns of {pca_prefix}.eigenvec file"

    fig, axes = plt.subplots(figsize=figsize, dpi=150, ncols=2)
    admix.plot.joint_pca(
        df_pc=df_plot,
        eigenval=eigenval,
        axes=axes,
        sample_alpha=sample_alpha,
        x=x,
        y=y,
    )
    fig.tight_layout()
    fig.savefig(f"{out}.png", bbox_inches="tight")
    admix.logger.info(f"PCA plots saved to {out}.png")



# flake8: noqa
from setuptools import setup, find_packages
from pathlib import Path

long_description = (Path(__file__).parent / "README.md").read_text()

exec(open("admix/version.py").read())

setup(
    name="admix-kit",
    version=__version__,
    description="Tool kits for analyzing genetics data of admixed population",
    url="https://github.com/kangchenghou/admix-kit",
    author="Kangcheng Hou",
    author_email="kangchenghou@gmail.com",
    packages=find_packages(),
    setup_requires=["numpy>=1.10"],
    python_requires=">=3.7",
    install_requires=[
        "Cython",
        "numpy",
        "pandas",
        "scipy",
        "matplotlib",
        "dask[array]>=2021.11.2",
        "tqdm",
        "xarray",
        "zarr",
        "statsmodels",
        "structlog",
        "fire",
        "pyreadr",
        "tinygwas",
        "dask-pgen",
        "haptools==0.4.0",
    ],
    entry_points={"console_scripts": ["admix=admix.cli:cli"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Science/Research",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
)



import pandas as pd
import xarray as xr
import numpy as np
import dask.array as da
from xarray.core.dataset import DataVariables
import admix
import dask
from typing import (
    Hashable,
    List,
    Optional,
    Any,
    Dict,
    Union,
    MutableMapping,
)
from ._index import normalize_indices
import warnings


class Dataset(object):
    """Data structure to contain genotype and local ancestry."""

    def __init__(
        self,
        geno: Optional[da.Array] = None,
        lanc: Optional[da.Array] = None,
        snp: Optional[pd.DataFrame] = None,
        indiv: Optional[pd.DataFrame] = None,
        n_anc: int = None,
        dset_ref=None,
        snp_idx: Union[slice, int, np.ndarray] = None,
        indiv_idx: Union[slice, int, np.ndarray] = None,
        enforce_order: bool = True,
    ):
        if dset_ref is not None:
            # initialize from reference data set
            if isinstance(snp_idx, (int, np.integer)):
                assert (
                    0 <= snp_idx < dset_ref.n_snp
                ), "SNP index `{snp_idx}` is out of range."
                snp_idx = slice(snp_idx, snp_idx + 1, 1)

            if isinstance(indiv_idx, (int, np.integer)):
                assert (
                    0 <= indiv_idx < dset_ref.n_indiv
                ), "SNP index `{indiv_idx}` is out of range."
                indiv_idx = slice(indiv_idx, indiv_idx + 1, 1)

            for name, idx in zip(["snp", "indiv"], [snp_idx, indiv_idx]):
                if isinstance(idx, slice):
                    assert (
                        (idx.start is None)
                        or (idx.stop is None)
                        or (idx.start < idx.stop)
                    ), f"Slice `{idx}` is not ordered."
                    if idx.step is not None:
                        assert idx.step > 0, f"Slice `{idx}` is not ordered."
                elif isinstance(idx, np.ndarray):
                    if enforce_order:
                        assert np.all(
                            idx == np.sort(idx)
                        ), f"idx=`{idx}` is not ordered according to dset.snp.index"
                else:
                    raise ValueError(
                        f"`{name}_idx` must be a slice or a numpy array of integers."
                    )
                if name == "snp":
                    df_subset = dset_ref.snp.iloc[idx, :].copy()
                    self._snp = df_subset
                elif name == "indiv":
                    df_subset = dset_ref.indiv.iloc[idx, :].copy()
                    self._indiv = df_subset
                else:
                    raise ValueError(f"Unknown name {name}")

            with dask.config.set(**{"array.slicing.split_large_chunks": False}):
                self._xr = dset_ref.xr.isel(snp=snp_idx, indiv=indiv_idx)

        else:
            # initialize from actual data set
            # assign `geno` and `lanc`
            assert geno is not None, "`geno` must not be None"
            data_vars: Dict[Hashable, Any] = {}
            data_vars["geno"] = (("snp", "indiv", "ploidy"), geno)

            n_snp, n_indiv = geno.shape[0:2]
            if lanc is not None:
                assert geno.shape == lanc.shape
                assert isinstance(lanc, da.Array), "`lanc` must be a dask array"
                data_vars["lanc"] = (("snp", "indiv", "ploidy"), lanc)

            # assign `indiv` and `snp`
            if snp is None:
                self._snp = pd.DataFrame(index=pd.RangeIndex(stop=n_snp))
            else:
                self._snp = snp

            if indiv is None:
                self._indiv = pd.DataFrame(index=pd.RangeIndex(stop=n_indiv))
            else:
                self._indiv = indiv

            self._xr = xr.Dataset(
                data_vars=data_vars,
                coords={"snp": self._snp.index, "indiv": self._indiv.index},
            )

            # assign attributes
            if ("lanc" in self._xr) and (n_anc is None):
                # infer number of ancestors
                n_anc = int(lanc[0:1000, :, :].max().compute() + 1)
                self._xr.attrs["n_anc"] = n_anc
            else:
                self._xr.attrs["n_anc"] = n_anc

            self._xr.attrs["path"] = None
        # self._check_dimensions()
        # self._check_uniqueness()

    def _check_dimensions(self) -> None:
        """
        Check that all dimensions are unique.
        """
        pass

    def _check_uniqueness(self) -> None:
        """
        Check that all dimensions are unique.
        """
        pass

    def __repr__(self) -> str:
        descr = (
            f"admix.Dataset object with n_snp x n_indiv = {self.n_snp} x {self.n_indiv}"
        )
        if "lanc" in self._xr:
            descr += f", n_anc={self.n_anc}"
        else:
            descr += ", no local ancestry"

        if len(self.snp.columns) > 0:
            descr += "\n\tsnp: " + ", ".join([f"'{col}'" for col in self.snp.columns])
        if len(self.indiv.columns) > 0:
            descr += "\n\tindiv: " + ", ".join(
                [f"'{col}'" for col in self.indiv.columns]
            )

        return descr

    @property
    def n_indiv(self) -> int:
        """Number of individuals."""
        return self._xr.dims["indiv"]

    @property
    def n_snp(self) -> int:
        """Number of SNPs."""
        return self._xr.dims["snp"]

    @property
    def n_anc(self) -> int:
        """Number of ancestries."""
        return self._xr.attrs["n_anc"]

    @property
    def data(self) -> DataVariables:
        """Number of individuals."""
        return self._xr.data_vars

    @property
    def indiv(self) -> pd.DataFrame:
        """One-dimensional annotation of observations (`pd.DataFrame`)."""
        return self._indiv

    @property
    def snp(self) -> pd.DataFrame:
        """One-dimensional annotation of observations (`pd.DataFrame`)."""
        return self._snp

    @property
    def geno(self) -> da.Array:
        """Genotype matrix"""
        return self._xr["geno"].data

    @property
    def lanc(self) -> da.Array:
        """Local ancestry matrix"""
        return self._xr["lanc"].data

    @property
    def xr(self) -> xr.Dataset:
        """Return the xr.Dataset used internally"""
        return self._xr

    def allele_per_anc(self) -> da.Array:
        """Return the allele-per-ancestry raw count matrix"""

        assert "lanc" in self._xr, "Local ancestry information is not available."
        return admix.data.allele_per_anc(
            geno=self.geno, lanc=self.lanc, n_anc=self.n_anc, center=False
        )

    def af_per_anc(self, force=False) -> da.Array:
        """
        Return the allele frequency per ancestry (n_snp, n_anc)

        nhaplo_per_anc will also be computed / updated.
        Parameters
        ----------
        force : bool
            If True, force re-computation of the matrix.
        """
        if ("af_per_anc" not in self._xr) or force:
            res = admix.data.af_per_anc(
                geno=self.geno,
                lanc=self.lanc,
                n_anc=self.n_anc,
                return_nhaplo=True,
            )
            self._xr["af_per_anc"] = ("snp", "anc"), res[0]
            self._xr["nhaplo_per_anc"] = ("snp", "anc"), res[1]

        return self._xr["af_per_anc"].data

    def nhaplo_per_anc(self, force=False) -> da.Array:
        """
        Return the number of haplotype per ancestry (n_snp, n_anc)

        af_per_anc will also be computed / updated.

        Parameters
        ----------
        force : bool
            If True, force re-computation of the matrix.
        """
        if ("nhaplo_per_anc" not in self._xr) or force:
            res = admix.data.af_per_anc(
                geno=self.geno,
                lanc=self.lanc,
                n_anc=self.n_anc,
                return_lanc_count=True,
            )
            self._xr["af_per_anc"] = ("snp", "anc"), res[0]
            self._xr["nhaplo_per_anc"] = ("snp", "anc"), res[1]

        return self._xr["nhaplo_per_anc"].data

    @property
    def uns(self) -> MutableMapping:
        """Unstructured annotation (ordered dictionary)."""
        uns = self._xr.attrs.get("uns", None)
        return uns

    def persist(self):
        """persist the lazy data to memory"""
        for name in ["geno", "lanc"]:
            if name in self._xr:
                self._xr[name] = (
                    self._xr[name].dims,
                    da.from_array(self._xr[name].data.compute(), chunks=-1),
                )

    def append_indiv_info(
        self, df_info: pd.DataFrame, force_update: bool = False
    ) -> None:
        """
        append indiv info to the dataset, individual is matched using the self.indiv.index
        and df_info.index. Missing individuals in df_info will be filled with NaN.

        Parameters
        ----------
        df_info : pd.DataFrame
            DataFrame with the indiv info
        force_update : bool
            If True, update the indiv information even if it already exists.
        """
        if len(set(df_info.index) - set(self.indiv.index)) > 0:
            admix.logger.warn(
                "admix.dataset.append_indiv_info: "
                f"{len(set(df_info.index) - set(self.indiv.index))}/{len(set(df_info.index))}"
                " individuals in the new dataframe not in the dataset;"
                " These individuals will be ignored."
            )
        if len(set(self.indiv.index) - set(df_info.index)) > 0:
            admix.logger.warn(
                "admix.dataset.append_indiv_info: "
                f"{len(set(self.indiv.index) - set(df_info.index))}/{len(set(self.indiv.index))}"
                " individuals in the dataset are missing in the provided data frame."
                " These individuals will be filled with NaN."
            )

        df_info = df_info.reindex(self.indiv.index)

        # for every column in df_info, if it is not in the dataset, add it to the dataset
        # else, check the consistency of between df_info and dataset, if not consistent, raise error
        for col in df_info.columns:
            if col in self.indiv.columns:
                is_allclose = np.allclose(self.indiv[col], df_info[col], equal_nan=True)
                if not is_allclose:
                    if force_update:
                        admix.logger.info(
                            f"admix.dataset.append_indiv_info: "
                            f"{col} is updated from {self.indiv[col].values[0:5]} ..."
                            f"to {df_info[col].values[0:5]} ..."
                        )
                        self._indiv[col] = df_info[col]
                    else:
                        raise ValueError(
                            "admix.dataset.append_indiv_info: "
                            f"The column '{col}' in the provided data frame is not consistent "
                            "with the dataset."
                        )
            else:
                self._indiv[col] = df_info[col]

    def append_snp_info(self, df_info: pd.DataFrame) -> None:
        """
        append snp info to the dataset, snp is matched using the self.snp.index
        and df_info.index.

        Parameters
        ----------
        df_info : pd.DataFrame
            DataFrame with the snp info
        """
        if len(set(df_info.index) - set(self.snp.index)) > 0:
            # raise warning
            warnings.warn("Some SNPs in the `df_info` are not in the dataset.")

        df_info = df_info.reindex(self.snp.index)

        # for every column in df_info, if it is not in the dataset, add it to the dataset
        # else, check the consistency of between df_info and dataset, if not consistent, raise error
        for col in df_info.columns:
            if col in self.snp.columns:
                assert np.allclose(
                    self.snp[col], df_info[col]
                ), f"The column {col} in the `df_info` is not consistent with the dataset."
            else:
                self._snp[col] = df_info[col]

    def __getitem__(self, index) -> "Dataset":
        """Returns a sliced view of the object."""
        snp_idx, indiv_idx = normalize_indices(index, self.snp.index, self.indiv.index)
        return Dataset(dset_ref=self, snp_idx=snp_idx, indiv_idx=indiv_idx)


def is_aligned(dset_list: List[Dataset], dim="snp"):
    """Check whether the datasets align with each other.

    Parameters
    ----------
    dset_list : List[Dataset]
        List of datasets to check
    dim: str
        Dimension to check. Either "snp" or "indiv"
    """
    if len(dset_list) == 0:
        return

    assert dim in ["snp", "indiv"], "dim must be either 'snp' or 'indiv'"

    if dim == "snp":
        df_snp_list = [dset.snp[["CHROM", "POS", "REF", "ALT"]] for dset in dset_list]
        return np.all([df_snp.equals(df_snp_list[0]) for df_snp in df_snp_list[1:]])
    elif dim == "indiv":
        df_indiv_list = [dset.indiv for dset in dset_list]
        return np.all(
            [df_indiv.equals(df_indiv_list[0]) for df_indiv in df_indiv_list[1:]]
        )
    else:
        raise ValueError("dim must be either 'snp' or 'indiv'")



"""
Standalone functions corresponding to the external tools
"""

import admix
import os
import subprocess
import pandas as pd
import shutil
from typing import List
import numpy as np
import dapgen
from . import get_dependency, get_cache_data


def hapgen2(
    pfile: str,
    n_indiv: int,
    out_prefix: str,
    genetic_map: str,
    chrom: int = None,
    plink_kwargs: dict = dict(),
):
    """simulate genotype with HAPGEN2

    Parameters
    ----------
    pfile : str
        path to PLINK2 pfile
    genetic_map : str
        genetic map, downloaded from
        https://alkesgroup.broadinstitute.org/Eagle/downloads/tables/
    chrom : int
        chromosome number for the plink2 file. The file can only contain data from one chromosome,
        if not specified, the function will try to guess the chromosome number from the plink2 file
    n_indiv : int
        Number of individuals to simulate
    plink_kwargs : dict
        kwargs to pass to plink2, e.g. --from-bp XX --to-bp YY
    out_prefix : str
        output prefix
    """
    tmp_dir = out_prefix + ".hapgen2tmpdata"
    assert not os.path.exists(
        tmp_dir
    ), f"{tmp_dir} should not exist, please remove it before running this function"
    os.makedirs(tmp_dir, exist_ok=False)
    tmp_data_prefix = os.path.join(tmp_dir, "hapgen2_data")

    # fetch chromosome
    if chrom is None:
        chrom = dapgen.read_pvar(pfile + ".pvar")["CHROM"].values
        assert (
            len(np.unique(chrom)) == 1
        ), "only one chromosome is allowed in the plink2 file"
        chrom = chrom[0]

    ##################################
    # convert PLINK2 to legend and hap
    ##################################

    admix.tools.plink2.run(
        " ".join(
            [
                "--pfile",
                pfile,
                "--export hapslegend",
                f"--out {tmp_data_prefix}",
                f"--chr {chrom}",
            ]
        ),
        **plink_kwargs,
    )

    ##################################
    # simulate genotype with HAPGEN2
    ##################################

    assert genetic_map in [
        "hg19",
        "hg38",
    ], f"genetic map {genetic_map} not supported, only hg19 and hg38 are supported"

    df_map = pd.read_csv(
        get_cache_data("genetic_map", build=genetic_map), delim_whitespace=True
    )
    df_map = df_map[df_map.chr == chrom].drop(columns=["chr"])
    df_map.to_csv(f"{tmp_data_prefix}.genetic_map", sep="\t", index=False)

    df_legend = pd.read_csv(f"{tmp_data_prefix}.legend", delim_whitespace=True, nrows=1)
    dl = df_legend["position"][0]

    hapgen2 = get_dependency("hapgen2")

    cmd_hapgen2 = [
        f"{hapgen2}",
        f" -m {tmp_data_prefix}.genetic_map",
        f" -l {tmp_data_prefix}.legend",
        f" -h {tmp_data_prefix}.haps",
        f" -o {tmp_data_prefix}.hapgen2.gz",
        f" -dl {dl} 1 1 1",
        f" -n {n_indiv} 0",
        " -no_gens_output",
    ]
    subprocess.check_call(" ".join(cmd_hapgen2), shell=True)

    ##################################
    # convert HAPGEN2 to PLINK2
    ##################################

    cmd_convert_pgen = [
        f"--haps {tmp_data_prefix}.hapgen2.controls.haps.gz",
        f"--legend {tmp_data_prefix}.hapgen2.legend {chrom}",
        "--make-pgen",
        f"--out {out_prefix}",
    ]

    admix.tools.plink2.run(" ".join(cmd_convert_pgen))

    shutil.move(f"{tmp_data_prefix}.hapgen2.gz.summary", f"{out_prefix}.hapgen2.log")

    ##################################
    # clean up
    ##################################
    shutil.rmtree(tmp_dir)


def interpolate_genetic_position(chrom: int, pos: np.ndarray, build: str) -> np.ndarray:
    """Intepolate physical position (bp) to genetic position (cM)

    Parameters
    ----------
    chrom : int
        which chromosome is the position `pos` on. Pass SNP position one chromosome
        at a time.
    pos : np.ndarray
        position
    build : str
        genome build

    Returns
    -------
    np.ndarray
        genetic position

    Notes
    -----
    When a given position is out of the range of the genetic map, the genetic position
    will be extrapolated linearly.
    """
    assert build in ["hg19", "hg38"]
    from scipy.interpolate import interp1d

    df_map = pd.read_csv(
        admix.tools.get_cache_data("genetic_map", build=build), delim_whitespace=True
    )
    df_map = df_map[df_map["chr"] == chrom]
    assert np.all(np.sort(df_map["position"]) == df_map["position"])
    interp = interp1d(
        x=df_map["position"],
        y=df_map["Genetic_Map(cM)"],
        assume_sorted=True,
        fill_value="extrapolate",
    )
    genetic_pos = interp(pos)
    return genetic_pos


def admix_simu(
    pfile_list: List[str],
    admix_prop: List[float],
    n_gen: int,
    n_indiv: int,
    build: str,
    out_prefix: str,
):
    """
    Wrapper for https://github.com/williamslab/admix-simu

    Parameters
    ----------
    pfile_list : List[str]
        list of input prefixes
    admix_prop :
        proportion of admixed individuals relative proportion of the admixture,
        will be rescaled to 1.
    n_gen : int
        number of generations to simulate
    n_indiv : int
        number of individuals to simulate
    out_prefix: str
        output prefix
    admix_simu_dir: str
        Directory to the admix_simu software

    Returns
    -------

    haps format
    -----------
    https://mathgen.stats.ox.ac.uk/genetics_software/shapeit/shapeit.html#hapsample

    phgeno format
    -------------
    <i1_hap1><i1_hap2><i2_hap1><i2_hap2>

    Steps
    -----
    1. Convert to phgeno format
    2. Run admix-simu
    3. Interpolate genetic position https://privefl.github.io/bigsnpr/reference/snp_asGeneticPos.html

    """

    ##################################
    # check input
    ##################################
    assert build in ["hg19", "hg38"], "build should be hg19 or hg38"
    assert (
        n_gen > 1
    ), "n_gen should be greater than 1, otherwise admix-simu will run with errors"
    assert len(pfile_list) == len(
        admix_prop
    ), "`pfile_list`, `admix_prop` should have the same length"
    n_pop = len(pfile_list)

    for i, pfile in enumerate(pfile_list):
        admix.logger.info(f"POP{i}={pfile}")

    admix_simu_dir = admix.tools.get_dependency("admix-simu")
    ##################################
    # format input
    ##################################
    tmp_dir = out_prefix + ".admixsimutmpdata"
    assert not os.path.exists(
        tmp_dir
    ), f"{tmp_dir} should not exist, please remove it before running this function"
    os.makedirs(tmp_dir, exist_ok=False)
    df_snp_info = None
    chrom = None
    for pfile_path in pfile_list:
        pfile = os.path.basename(pfile_path)
        admix.tools.plink2.run(
            f"--pfile {pfile_path} --export hapslegend --out {tmp_dir}/{pfile}"
        )
        # convert haps to phgeno by removing all spaces
        subprocess.check_output(
            f"cat {tmp_dir}/{pfile}.haps | tr -d ' ' > {tmp_dir}/{pfile}.phgeno",
            shell=True,
        )
        # check all legend files are the same
        if df_snp_info is None:
            df_snp_info = pd.read_csv(
                f"{tmp_dir}/{pfile}.legend", delim_whitespace=True
            )
        else:
            assert df_snp_info.equals(
                pd.read_csv(f"{tmp_dir}/{pfile}.legend", delim_whitespace=True)
            ), f"SNP information in {pfile} are not the same as in {pfile_list[0]}"

        # check all chromosomes are the same and only one chromosome
        tmp_chrom = dapgen.read_pvar(f"{pfile_path}.pvar")["CHROM"]
        assert len(set(tmp_chrom)) == 1, f"Multiple chromosomes in {pfile_path}.pvar"
        if chrom is None:
            chrom = tmp_chrom[0]
        else:
            assert (
                chrom == tmp_chrom[0]
            ), f"Chromosomes in {pfile_path}.pvar are not the same as in {pfile_list[0]}"

    ##################################
    # run admix-simu
    ##################################
    phgeno_list = [
        f"-POP{i + 1} {tmp_dir}/{os.path.basename(pfile)}.phgeno"
        for i, pfile in enumerate(pfile_list)
    ]

    admix_dat = [
        "\t".join(
            [str(n_indiv * 2), "ADMIX", *[f"POP{i}" for i in np.arange(1, n_pop + 1)]]
        ),
        "\t".join([str(n_gen), "0", *[str(prop) for prop in admix_prop]]),
    ]

    dat_file = os.path.join(tmp_dir, "admix.dat")
    with open(dat_file, "w") as f:
        f.writelines("\n".join(admix_dat))

    # cat dat_file
    print("\n".join(admix_dat))
    df_snp_info.insert(1, "chrom", chrom)
    # insert Morgan genetic position (cM / 100)
    df_snp_info.insert(
        2,
        "M",
        interpolate_genetic_position(
            chrom=chrom, pos=df_snp_info["position"], build=build
        )
        / 100,
    )
    snp_file = os.path.join(tmp_dir, "snp_info.txt")
    df_snp_info.to_csv(snp_file, sep="\t", index=False, header=False)

    cmd = " ".join(
        [
            os.path.join(admix_simu_dir, "simu-mix.pl"),
            dat_file,
            snp_file,
            os.path.join(tmp_dir, "admix"),
            *phgeno_list,
        ]
    )
    admix.logger.info(cmd)
    subprocess.check_output(cmd, shell=True)

    cmd = " ".join(
        [
            os.path.join(admix_simu_dir, "bp2anc.pl"),
            os.path.join(tmp_dir, "admix.bp"),
            ">",
            os.path.join(tmp_dir, "admix.hanc"),
        ]
    )
    admix.logger.info(cmd)
    subprocess.check_output(cmd, shell=True)

    ##################################
    # post-processing
    ##################################

    ## convert phgeno to pgen
    df_snp_info[["id", "position", "a0", "a1"]].to_csv(
        os.path.join(tmp_dir, "admix.legend"), sep=" ", index=False
    )
    subprocess.check_output(
        f"cat {tmp_dir}/admix.phgeno | sed 's/./& /g' > {tmp_dir}/admix.haps",
        shell=True,
    )
    admix.tools.plink2.run(
        f"--haps {tmp_dir}/admix.haps --legend {tmp_dir}/admix.legend {chrom} "
        + f"--make-pgen --out {tmp_dir}/admix",
    )
    for suffix in ["pgen", "pvar", "psam", "log"]:
        shutil.move(os.path.join(tmp_dir, f"admix.{suffix}"), f"{out_prefix}.{suffix}")

    ## convert local ancestry
    lanc = admix.data.read_bp_lanc(os.path.join(tmp_dir, "admix.bp"))
    lanc.write(f"{out_prefix}.lanc")
    shutil.move(os.path.join(tmp_dir, "admix.hanc"), f"{out_prefix}.hanc")
    shutil.move(os.path.join(tmp_dir, "admix.bp"), f"{out_prefix}.bp")

    # clean up
    shutil.rmtree(tmp_dir)

    # hint the file path
    admix.logger.info(f"Output files are saved to {out_prefix}.*")


def haptools_simu_admix(
    pfile: str,
    admix_prop: List[float],
    pop_col: str,
    mapdir: str,
    n_gen: int,
    n_indiv: int,
    out_prefix: str,
    seed=1234,
):
    """Wrapper for haptools simgenotype

    Parameters
    ----------
    pfile : str
        list of input prefixes
    admix_prop :
        proportion of admixed individuals relative proportion of the admixture,
        will be rescaled to 1.
    n_gen : int
        number of generations to simulate
    n_indiv : int
        number of individuals to simulate
    out_prefix: str
        output prefix
    """
    from bisect import bisect_left

    ##################################
    # check input
    ##################################
    assert (
        n_gen > 1
    ), "n_gen should be greater than 1, otherwise admix-simu will run with errors"

    ##################################
    # format input
    ##################################
    tmp_dir = out_prefix + ".tmpdata"
    assert not os.path.exists(
        tmp_dir
    ), f"{tmp_dir} should not exist, please remove it before running this function"
    os.makedirs(tmp_dir, exist_ok=False)

    ##################################
    # run haptools
    ##################################

    admix_dat = [
        "\t".join([str(n_indiv), "ADMIX", *[pop for pop in admix_prop]]),
        "\t".join([str(n_gen), "0", *[str(admix_prop[pop]) for pop in admix_prop]]),
    ]

    dat_file = os.path.join(tmp_dir, "admix.dat")
    with open(dat_file, "w") as f:
        f.writelines("\n".join(admix_dat))

    # cat dat_file
    print("\n".join(admix_dat))

    sample_df = dapgen.read_psam(pfile + ".psam")
    sample_info_file = os.path.join(tmp_dir, "sample_info.txt")
    sample_df[[pop_col]].to_csv(sample_info_file, sep="\t", index=True, header=True)

    snp_df = dapgen.read_pvar(pfile + ".pvar")
    chrom = np.unique(snp_df["CHROM"])
    assert len(chrom) == 1, "Only one chromosome is allowed in the plink2 file"
    start, stop = np.min(snp_df["POS"]), np.max(snp_df["POS"])

    cmds = [
        "haptools simgenotype",
        f"--model {dat_file}",
        f"--mapdir {mapdir}",
        f"--ref_vcf {pfile}.pgen",
        f"--sample_info {sample_info_file}",
        f"--out {tmp_dir}/admix.pgen",
        f"--region {chrom[0]}:{start}-{stop}",
        f"--seed {seed}",
    ]
    cmd = " ".join(cmds)
    admix.logger.info(cmd)
    subprocess.check_output(cmd, shell=True)

    ##################################
    # post-processing
    ##################################

    anc_map = {pop: order for order, pop in enumerate(admix_prop)}

    with open(os.path.join(tmp_dir, "admix.bp")) as f:
        lines = f.readlines()

    seps = list(np.where([line.startswith("Sample_") for line in lines])[0]) + [
        len(lines)
    ]

    breaks = []
    values = []
    for i in range(len(seps) - 1):
        haplo = lines[seps[i] + 1 : seps[i + 1]]

        hap_pos = []
        hap_values = []
        for line in haplo:
            anc, chrom, pos, cm = line.strip().split()
            anc_order = anc_map[anc]
            hap_pos.append(int(pos))
            hap_values.append(anc_order)
        hap_breaks = [bisect_left(snp_df.POS.values, p) for p in hap_pos]
        breaks.append(hap_breaks)
        values.append(hap_values)

    dip_breaks, dip_values = admix.data.clean_lanc(
        *admix.data.haplo2diplo(breaks=breaks, values=values),
        remove_repeated_val=True,
    )
    lanc = admix.data.Lanc(breaks=dip_breaks, values=dip_values)
    lanc.write(f"{out_prefix}.lanc")

    for suffix in ["pgen", "pvar", "psam", "bp"]:
        shutil.move(os.path.join(tmp_dir, f"admix.{suffix}"), f"{out_prefix}.{suffix}")

    # clean up
    shutil.rmtree(tmp_dir)

    # hint the file path
    admix.logger.info(f"Output files are saved to {out_prefix}.*")



import pandas as pd
import numpy as np
from typing import (
    List,
    Union,
    Tuple,
    Sequence,
)


def normalize_indices(
    index, snp_names: pd.Index, indiv_names: pd.Index
) -> Tuple[Union[slice, int, np.ndarray], Union[slice, int, np.ndarray]]:
    """Normalize the indices to return the snp slices, and individual slices

    Parameters
    ----------
    index : [type]
        [description]
    snp_names : pd.index
        [description]
    indiv_names : pd.index
        [description]

    Returns
    -------
    [type]
        [description]

    Raises
    ------
    ValueError
        [description]
    """
    # deal with tuples of length 1
    if isinstance(index, tuple) and len(index) == 1:
        index = index[0]

    if isinstance(index, tuple):
        if len(index) > 2:
            raise ValueError(
                "data can only be sliced in SNPs (first dim) and individuals (second dim)"
            )

    snp_ax, indiv_ax = unpack_index(index)
    snp_ax = _normalize_index(snp_ax, snp_names)
    indiv_ax = _normalize_index(indiv_ax, indiv_names)
    return snp_ax, indiv_ax


# convert the indexer (integer, slice, string, array) to the actual positions
# reference: https://github.com/theislab/anndata/blob/566f8fe56f0dce52b7b3d0c96b51d22ea7498156/anndata/_core/index.py#L16
def _normalize_index(
    indexer: Union[
        slice,
        int,
        str,
        np.ndarray,
    ],
    index: pd.Index,
) -> Union[slice, int, np.ndarray]:  # ndarray of int
    """Convert the indexed (integer, slice, string, array) to the actual positions

    Parameters
    ----------
    indexer : Union[ slice, int, str, np.ndarray, ]
        [description]
    index : pd.Index
        [description]

    Returns
    -------
    Union[slice, int, np.ndarray]
        [description]

    """
    if not isinstance(index, pd.RangeIndex):
        assert (
            index.dtype != float and index.dtype != int
        ), "Don't call _normalize_index with non-categorical/string names"

    # the following is insanely slow for sequences,
    # we replaced it using pandas below
    def name_idx(i):
        if isinstance(i, str):
            i = index.get_loc(i)
        return i

    if isinstance(indexer, slice):
        start = name_idx(indexer.start)
        stop = name_idx(indexer.stop)
        # string slices can only be inclusive, so +1 in that case
        if isinstance(indexer.stop, str):
            stop = None if stop is None else stop + 1
        step = indexer.step
        return slice(start, stop, step)
    elif isinstance(indexer, (np.integer, int)):
        return indexer
    elif isinstance(indexer, str):
        return index.get_loc(indexer)  # int
    elif isinstance(indexer, (Sequence, np.ndarray, pd.Index, np.matrix)):
        if hasattr(indexer, "shape") and (
            (indexer.shape == (index.shape[0], 1))
            or (indexer.shape == (1, index.shape[0]))
        ):
            indexer = np.ravel(indexer)
        if not isinstance(indexer, (np.ndarray, pd.Index)):
            indexer = np.array(indexer)
        if issubclass(indexer.dtype.type, (np.integer, np.floating)):
            return indexer  # Might not work for range indexes
        elif issubclass(indexer.dtype.type, np.bool_):
            if indexer.shape != index.shape:
                raise IndexError(
                    f"Boolean index does not match Dataset's shape along this "
                    f"dimension. Boolean index has shape {indexer.shape} while "
                    f"Dataset index has shape {index.shape}."
                )
            positions = np.where(indexer)[0]
            return positions  # np.ndarray[int]
        else:  # indexer should be string array
            positions = index.get_indexer(indexer)
            if np.any(positions < 0):
                not_found = indexer[positions < 0]
                raise KeyError(
                    f"Values {list(not_found)}, from {list(indexer)}, "
                    "are not valid obs/ var names or indices."
                )
            return positions  # np.ndarray[int]
    else:
        raise IndexError(f"Unknown indexer {indexer!r} of type {type(indexer)}")


def unpack_index(index):
    if not isinstance(index, tuple):
        return index, slice(None)
    elif len(index) == 2:
        return index
    elif len(index) == 1:
        return index[0], slice(None)
    else:
        raise IndexError("invalid number of indices")



import admix
import os
import subprocess
import glob
import shutil
import pandas as pd
import numpy as np
import dapgen
from typing import Tuple


def log_params(name, params):
    admix.logger.info(
        f"Received parameters: \n{name}\n  "
        + "\n  ".join(f"--{k}={v}" for k, v in params.items())
    )


def _process_sample_map(root_dir):
    """Download and format sample map from 1000 Genomes"""

    sample_map = pd.read_csv(
        f"{root_dir}/pgen/all_chr.psam",
        delim_whitespace=True,
    )
    unrelated_id = pd.read_csv(
        f"{root_dir}/pgen/king.cutoff.out.id", delim_whitespace=True
    )
    os.makedirs(f"{root_dir}/metadata", exist_ok=True)
    sample_map[["#IID", "Population", "SuperPop"]].to_csv(
        f"{root_dir}/metadata/full_sample.tsv",
        sep="\t",
        index=False,
        header=False,
    )
    # filter unrelated
    unrelated_sample_map = sample_map[~sample_map["#IID"].isin(unrelated_id["#IID"])]
    unrelated_sample_map[["#IID", "Population", "SuperPop"]].to_csv(
        f"{root_dir}/metadata/unrelated_sample.tsv",
        sep="\t",
        index=False,
        header=False,
    )
    admix.logger.info("Population in unrelated sample map:")
    admix.logger.info(unrelated_sample_map["Population"].value_counts())


def _process_genetic_map(root_dir, build):
    """
    Download and format genetic map
    1. call bash script to download genetic map to out/metadata/genetic_map/raw
    2. process the genetic map and save to out/metadata/genetic_map
    """

    assert build in ["hg19", "hg38"], "build should be hg38 or hg19"

    if build == "hg38":
        name = "GRCh38"
    elif build == "hg19":
        name = "GRCh37"
    else:
        raise ValueError("build should be hg38 or hg19")

    cmds = f"""
        mkdir -p {root_dir}/metadata/genetic_map/raw && cd {root_dir}/metadata/genetic_map/raw
        wget https://bochet.gcc.biostat.washington.edu/beagle/genetic_maps/plink.{name}.map.zip
        unzip plink.{name}.map.zip
    """

    subprocess.check_output(cmds, shell=True)

    for chrom in range(1, 23):
        # move raw files to out/metadata/genetic_map
        shutil.copy(
            f"{root_dir}/metadata/genetic_map/raw/plink.chr{chrom}.{name}.map",
            f"{root_dir}/metadata/genetic_map/chr{chrom}.map",
        )
    # clean up
    shutil.rmtree(f"{root_dir}/metadata/genetic_map/raw")

    if False:
        # OLD Eagle2 genetic map
        raw_map_path = admix.tools.get_cache_data("genetic_map", build=build)

        raw_map = pd.read_csv(
            raw_map_path,
            delim_whitespace=True,
        )

        os.makedirs(f"{root_dir}/metadata/genetic_map", exist_ok=True)
        for chrom in range(1, 23):
            chrom_map = raw_map[raw_map["chr"] == chrom].iloc[:, [0, 1, 3]]
            chrom_map.to_csv(
                f"{root_dir}/metadata/genetic_map/chr{chrom}.tsv",
                sep="\t",
                index=False,
                header=False,
            )


def get_1kg_ref(
    dir: str,
    build: str,
    verbose: bool = False,
    memory: int = 16000,
):
    """
    Get the 1,000 reference genome in plink2 format.

    Parameters
    ----------
    dir : str
        Directory where the reference genome is stored. The following files will be downloaded
        and stored in this directory:
        - ${dir}/pgen: containing plink2 files
        - ${dir}/metadata: containing metadata files
    build : str
        Build of the reference genome. hg38 or hg19.
    verbose : bool
        Whether to print out the progress of the download.
    memory : int
        Memory in MB to use for PLINK command. Default is 16000 (~16GB).

    Examples
    --------
    .. code-block:: bash

        admix get-1kg-ref --dir=/path/to/1kg --build hg38

    """
    log_params("get-1kg-ref", locals())

    def call_helper(cmd):
        admix.logger.info(f"$ {cmd}")
        if verbose:
            subprocess.run(cmd, shell=True, check=True)
        else:
            subprocess.run(
                cmd,
                shell=True,
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT,
            )

    # make sure plink2 is in path
    plink2 = admix.tools.get_dependency("plink2")

    # step0: download metadata
    os.makedirs(os.path.join(dir, "metadata"), exist_ok=True)
    admix.logger.info("Retrieving meta data...")

    # step1: download pgen
    os.makedirs(os.path.join(dir, "pgen"), exist_ok=True)
    if build == "hg38":
        cmds = [
            f"mkdir -p {dir}/pgen &&",
            f"wget https://www.dropbox.com/s/j72j6uciq5zuzii/all_hg38.pgen.zst?dl=1 -O {dir}/pgen/raw.pgen.zst &&",
            f"wget https://www.dropbox.com/s/ngbo2xm5ojw9koy/all_hg38_noannot.pvar.zst?dl=1 -O {dir}/pgen/raw.pvar.zst &&",
            f"wget https://www.dropbox.com/s/2e87z6nc4qexjjm/hg38_corrected.psam?dl=1 -O {dir}/pgen/raw.psam &&",
            f"wget https://www.dropbox.com/s/4zhmxpk5oclfplp/deg2_hg38.king.cutoff.out.id?dl=1 -O {dir}/pgen/king.cutoff.out.id",
        ]
    elif build == "hg19":
        cmds = [
            f"mkdir -p {dir}/pgen &&",
            f"wget https://www.dropbox.com/s/dps1kvlq338ukz8/all_phase3_ns.pgen.zst?dl=1 -O {dir}/pgen/raw.pgen.zst &&",
            f"wget https://www.dropbox.com/s/uqk3gfhwsvf7bf3/all_phase3_ns_noannot.pvar.zst?dl=1 -O {dir}/pgen/raw.pvar.zst &&",
            f"wget https://www.dropbox.com/s/6ppo144ikdzery5/phase3_corrected.psam?dl=1 -O {dir}/pgen/raw.psam &&",
            f"wget https://www.dropbox.com/s/zj8d14vv9mp6x3c/deg2_phase3.king.cutoff.out.id?dl=1 -O {dir}/pgen/king.cutoff.out.id",
        ]
    else:
        raise ValueError(f"Unknown build: {build}")

    # assume that ${dir}/pgen already contains raw.pgen / raw.pvar / raw.psam / king.cutoff.out.id
    if not all(
        [
            os.path.exists(f"{dir}/pgen/{f}")
            for f in ["raw.pgen.zst", "raw.pvar.zst", "raw.psam"]
        ]
    ):
        admix.logger.info(
            f"Please download the reference genome to {dir}/pgen using the following commands:\n"
            + "$ "
            + " ".join(cmds)
        )
        return

    admix.logger.info("Decompressing plink files")
    cmds = [
        f"{plink2} --zst-decompress {dir}/pgen/raw.pgen.zst > {dir}/pgen/raw.pgen &&",
        f"{plink2} --zst-decompress {dir}/pgen/raw.pvar.zst > {dir}/pgen/raw.pvar",
    ]
    call_helper(" ".join(cmds))

    admix.logger.info(
        "Basic QCing: bi-allelic SNPs, MAC >= 5, chromosome 1-22, unify SNP names"
    )
    cmds = [
        f"{plink2} --pfile {dir}/pgen/raw",
        "--allow-extra-chr",
        "--rm-dup exclude-all",
        "--max-alleles 2",
        "--mac 5",
        "--snps-only",
        "--chr 1-22",
        "--set-all-var-ids @:#:\$r:\$a",
        f"--make-pgen --out {dir}/pgen/raw2",
        f"--memory {memory}",
    ]
    call_helper(" ".join(cmds))

    ## remove SNPs duplicated by position
    df_pvar = dapgen.read_pvar(f"{dir}/pgen/raw2.pvar")
    dup_snps = df_pvar.index[df_pvar.duplicated(subset=["CHROM", "POS"])]
    admix.logger.info(
        f"Excluding {len(dup_snps)}/{df_pvar.shape[0]} SNPs duplicated by positions."
    )
    np.savetxt(f"{dir}/pgen/raw2.snplist", dup_snps, fmt="%s")
    cmds = [
        f"{plink2} --pfile {dir}/pgen/raw2",
        f"--exclude {dir}/pgen/raw2.snplist",
        f"--make-pgen --out {dir}/pgen/all_chr",
        f"--memory {memory}",
    ]
    call_helper(" ".join(cmds))

    admix.logger.info("Clean up temporary files...")
    # remove raw2*
    for f in glob.glob(f"{dir}/pgen/raw2*"):
        os.remove(f)

    _process_genetic_map(root_dir=dir, build=build)
    _process_sample_map(root_dir=dir)


def select_admix_indiv(
    ref_pfile: str,
    pca_prefix: str,
    superpop1: str,
    superpop2: str,
    out: str,
    exclude_pop1: str = None,
    exclude_pop2: str = None,
    n_pc: int = 4,
    sample_t_range: Tuple[float, float] = (0.05, 0.95),
    sample_dist_max: float = 1.5,
):
    """Select admixed individuals based on joint PCA analysis results of
    (1) reference dataset (2) sample dataset

    Parameters
    ----------
    ref_pfile : str
        reference panel pfile prefix
    pca_prefix : str
        joint pca results prefix. {pca_prefix}.eigenvec, {pca_prefix}.eigenval
        will be read
    out : str
        output prefix
    superpop1 : str
        superpopulation 1
    superpop2 : str
        superpopulation 2
    exclude_pop1 : str, optional
        exclude individuals from superpopulation 1, by default None
    exclude_pop2 : str, optional
        exclude individuals from superpopulation 2, by default None
    """
    log_params("select-admixed-indiv", locals())

    df_pc, eigenval = admix.io.read_joint_pca(
        ref_pfile=ref_pfile, pca_prefix=pca_prefix
    )
    pc_cols = [f"PC{i}" for i in range(1, n_pc + 1)]
    pc_col_pos = [df_pc.columns.get_loc(col) for col in pc_cols]
    pc_eigenval = eigenval[pc_col_pos]

    df_sample_pc = df_pc[df_pc.SUPERPOP == "SAMPLE"]
    df_anc1_pc = df_pc[df_pc.SUPERPOP == superpop1]
    df_anc2_pc = df_pc[df_pc.SUPERPOP == superpop2]

    if exclude_pop1 is not None:
        df_anc1_pc = df_anc1_pc[~df_anc1_pc.POP.isin(exclude_pop1)]
    if exclude_pop2 is not None:
        df_anc2_pc = df_anc2_pc[~df_anc2_pc.POP.isin(exclude_pop2)]

    df_sample_pc, df_anc1_pc, df_anc2_pc = (
        df_sample_pc[pc_cols],
        df_anc1_pc[pc_cols],
        df_anc2_pc[pc_cols],
    )

    # prepare sample_pc, anc1_pc, anc2_pc
    sample_dist, sample_t = admix.data.distance_to_refpop(
        df_sample_pc, df_anc1_pc, df_anc2_pc, weight=np.sqrt(pc_eigenval)
    )

    selected_mask = (
        (sample_t_range[0] < sample_t)
        & (sample_t < sample_t_range[1])
        & (sample_dist < sample_dist_max)
    )
    selected_indiv = df_sample_pc.index[selected_mask]
    admix.logger.info(
        f"{len(selected_indiv)}/{len(df_sample_pc)} selected to be admixed individuals."
    )

    df_plot = pd.concat([df_pc[df_pc.SUPERPOP != "SAMPLE"], df_pc.loc[selected_indiv]])
    import matplotlib.pyplot as plt

    fig, axes = plt.subplots(figsize=(8.5, 4), dpi=150, ncols=2)
    admix.plot.joint_pca(df_pc=df_plot, eigenval=eigenval, axes=axes)
    fig.tight_layout()
    fig.savefig(f"{out}.png", bbox_inches="tight")
    np.savetxt(f"{out}.indiv", selected_indiv, fmt="%s")
    admix.logger.info(f"PCA plots saved to {out}.png.")
    admix.logger.info(f"Selected individuals saved to {out}.indiv.")



import numpy as np
from typing import List, Tuple
from admix.tools import get_cache_data
import pandas as pd
from bisect import bisect_right, bisect_left


def calculate_mosaic_size(
    df_snp: pd.DataFrame, genetic_map: str, chrom: int, n_gen: int
) -> float:
    """Calculate the expected mosaic size in number of SNPs.

    1. Calculate the length of the region in cM using df_snp and genetic_map
    2. The expected number of cross-over events is calculated as
        n_cross_over = length_in_cM * n_gen
    3. The expected mosaic size is calculated as
        expected_mosaic_size = n_snp / n_cross_over

    Parameters
    ----------
    df_snp: pd.DataFrame
        DataFrame of SNP information
    genetic_map: str
        map to use, either 'hg19' or 'hg38'
    chrom: int
        specify the chromosome df_snp is, df_snp can only have one chromosome
    n_gen: int
        number of generations to simulate
    """

    assert np.all(df_snp.CHROM == chrom)
    assert genetic_map in ["hg19", "hg38"], "genetic_map must be either hg19 or hg38"

    n_snp = len(df_snp)

    # number of centimorgans
    df_map = pd.read_csv(
        get_cache_data("genetic_map", build=genetic_map), delim_whitespace=True
    )
    df_map = df_map[df_map.chr == chrom].drop(columns=["chr"])
    # find the closest SNP to df_snp.POS[0] and df_snp.POS[-1]
    # and calculate the length of the region in cM
    cm_start = df_map["Genetic_Map(cM)"].values[
        np.argmin(np.abs(df_map["position"] - df_snp.POS.values[0]))
    ]
    cm_stop = df_map["Genetic_Map(cM)"].values[
        np.argmin(np.abs(df_map["position"] - df_snp.POS.values[-1]))
    ]
    cm_length = cm_stop - cm_start

    # 1 cross-over per Morgan (centiMorgan / 100)
    expected_mosaic_size = n_snp / (n_gen * cm_length / 100)
    return expected_mosaic_size


def hap_lanc(
    n_snp: int,
    n_hap: int,
    mosaic_size: float,
    anc_props: List[float],
) -> Tuple[List[List[int]], List[List[int]]]:

    """Simulate local ancestries based on Poisson process. The simulated Poisson process
    will be homogeneous because non-homogeneous Poisson process is not easily supported.
    We will take the use df_snp to calculate the length of the region in cM. And assume
    a constant rate of cross-over.

    Haploid data are generated, use admix.data.haplo2diplo to combine pairs of
    haplotypes to get diploid data.

    The simulation process is as follows:
    1. Calculate the length of the region in cM using df_snp and genetic_map
    2. The expected number of cross-over events is calculated as
        n_cross_over = length_in_cM * n_gen
    3. The expected mosaic size is calculated as
        expected_mosaic_size = n_snp / n_cross_over


    Parameters
    ----------
    n_gen: int
        Number of generations to simulate
    n_hap: int
        Number of haplotypes to simulate
    mosaic_size: float
        Expected mosaic size in number of SNPs, use admix.simulate.calculate_mosaic_size
        to compute
    anc_props : list of float
        Proportion of ancestral populations, if not specified, the proportion
        is uniform over the ancestral populations.

    Returns
    -------
    np.ndarray
        Simulated local ancestry
    """
    assert np.sum(anc_props) == 1, "anc_props must sum to 1"

    n_total_snp = n_hap * n_snp

    # number of chunks to simulate in each iteration, simulate until
    # the desired total number of SNPs is reached
    chunk_size = int(n_total_snp / mosaic_size)

    raw_breaks: List[float] = []
    while np.sum(raw_breaks) < n_total_snp:
        raw_breaks.extend(np.random.exponential(scale=mosaic_size, size=chunk_size))
    breaks = np.cumsum(np.ceil(raw_breaks).astype(int))
    # find first break that is larger than the desired number of SNPs
    breaks = breaks[0 : int(np.argmax(breaks > n_total_snp) + 1)]
    # simulate local ancestry values
    values = np.random.choice(np.arange(len(anc_props)), size=len(breaks), p=anc_props)

    # insert values at n_snp, 2 * n_snp, ...
    boundary_loc = [
        bisect_right(breaks, v) for v in np.arange(1, n_hap + 1).astype(int) * n_snp
    ]

    # ...[:-1] to remove the last chunk because that would correspond to the extra
    # (n_hap + 1) haplotype
    break_list = np.split(breaks, boundary_loc)[:-1]
    value_list = np.split(values, boundary_loc)[:-1]
    # mod(..., n_snp + 1) to retain the boundary case of n_snp
    break_list = [np.mod(br, n_snp + 1).tolist() + [n_snp] for br in break_list]

    # the boundary take the value of next break
    boundary_values = values[boundary_loc]
    value_list = [vl.tolist() + [boundary_values[i]] for i, vl in enumerate(value_list)]

    # shuffle the break_list and value_list
    shuffled = list(zip(break_list, value_list))
    np.random.shuffle(shuffled)
    break_list, value_list = zip(*shuffled)

    return break_list, value_list